@@ -36,6 +36,8 @@
 #include <kimap/sessionuiproxy.h>
 
 #include <kimap/capabilitiesjob.h>
+#include <kimap/namespacejob.h>
+#include <kimap/listjob.h>
 #include <kimap/loginjob.h>
 #include <kimap/logoutjob.h>
 
@@ -53,7 +55,7 @@ class SessionUiProxy : public KIMAP::SessionUiProxy {
 };
 
 ImapAccount::ImapAccount( Settings *settings, QObject *parent )
-  : QObject( parent ), m_session( 0 ),
+  : QObject( parent ), m_mainSession( 0 ),
     m_encryption( KIMAP::LoginJob::Unencrypted ),
     m_authentication( KIMAP::LoginJob::ClearText ),
     m_subscriptionEnabled( false )
@@ -107,7 +109,7 @@ ImapAccount::ImapAccount( Settings *settings, QObject *parent )
 }
 
 ImapAccount::ImapAccount( QObject *parent )
-  : QObject( parent ), m_session( 0 ),
+  : QObject( parent ), m_mainSession( 0 ),
     m_encryption( KIMAP::LoginJob::Unencrypted ),
     m_authentication( KIMAP::LoginJob::ClearText ),
     m_subscriptionEnabled( false )
@@ -168,9 +170,9 @@ bool ImapAccount::isSubscriptionEnabled() const
   return m_subscriptionEnabled;
 }
 
-KIMAP::Session *ImapAccount::session() const
+KIMAP::Session *ImapAccount::mainSession() const
 {
-  return m_session;
+  return m_mainSession;
 }
 
 QStringList ImapAccount::capabilities() const
@@ -178,23 +180,52 @@ QStringList ImapAccount::capabilities() const
   return m_capabilities;
 }
 
+QList<KIMAP::MailBoxDescriptor> ImapAccount::namespaces() const
+{
+  return m_namespaces;
+}
+
 void ImapAccount::doConnect( const QString &password )
 {
   m_capabilities.clear();
+  m_namespaces.clear();
+  m_mainSession = createSessionInternal( password );
+}
 
-  if ( m_server.isEmpty() ) {
-    return;
+KIMAP::Session *ImapAccount::extraSession( const QString &id, const QString &password )
+{
+  if ( id.isEmpty() ) {
+    return 0;
   }
 
-  QString server = m_server.section( ":", 0, 0 );
-  int port = m_server.section( ":", 1, 1 ).toInt();
+  if ( m_extraSessions.contains( id ) ) {
+    return m_extraSessions[id];
+  }
+
+  KIMAP::Session *session = createSessionInternal( password );
+
+  if ( session!=0 ) {
+    m_extraSessions[id] = session;
+  }
+
+  return session;
+}
+
+KIMAP::Session *ImapAccount::createSessionInternal( const QString &password )
+{
+  if ( m_server.isEmpty() ) {
+    return 0;
+  }
+
+  QString server = m_server.section( ':', 0, 0 );
+  int port = m_server.section( ':', 1, 1 ).toInt();
 
   if ( m_encryption != KIMAP::LoginJob::Unencrypted && !QSslSocket::supportsSsl() ) {
     kWarning() << "Crypto not supported!";
-    emit error( EncryptionError,
+    emit error( 0, EncryptionError,
                 i18n( "You requested TLS/SSL to connect to %1, but your "
                       "system does not seem to be set up for that.", m_server ) );
-    return;
+    return 0;
   }
 
   if ( port <= 0 ) {
@@ -206,10 +237,10 @@ void ImapAccount::doConnect( const QString &password )
     }
   }
 
-  m_session = new KIMAP::Session( server, port, this );
-  m_session->setUiProxy( new SessionUiProxy );
+  KIMAP::Session *session = new KIMAP::Session( server, port, this );
+  session->setUiProxy( new SessionUiProxy );
 
-  KIMAP::LoginJob *loginJob = new KIMAP::LoginJob( m_session );
+  KIMAP::LoginJob *loginJob = new KIMAP::LoginJob( session );
   loginJob->setUserName( m_userName );
   loginJob->setPassword( password );
   loginJob->setEncryptionMode( m_encryption );
@@ -218,39 +249,64 @@ void ImapAccount::doConnect( const QString &password )
   QObject::connect( loginJob, SIGNAL( result( KJob* ) ),
                     this, SLOT( onLoginDone( KJob* ) ) );
   loginJob->start();
+
+  return session;
 }
 
 /******************* Slots  ***********************************************/
 
 void ImapAccount::onLoginDone( KJob *job )
 {
+  KIMAP::LoginJob *login = static_cast<KIMAP::LoginJob*>( job );
+
   if ( job->error()==0 ) {
-    KIMAP::CapabilitiesJob *capJob = new KIMAP::CapabilitiesJob( m_session );
+    if ( login->session()!=m_mainSession ) return; // No need to test for extra sessions
+    KIMAP::CapabilitiesJob *capJob = new KIMAP::CapabilitiesJob( m_mainSession );
     QObject::connect( capJob, SIGNAL( result( KJob* ) ), SLOT( onCapabilitiesTestDone( KJob* ) ) );
     capJob->start();
     return;
   } else {
-    emit error( LoginFailError,
+    emit error( login->session(), LoginFailError,
                 i18n( "Could not connect to the IMAP-server %1.", m_server ) );
-    disconnect();
+
+    QString id = m_extraSessions.key( login->session() );
+    if ( !id.isEmpty() ) {
+      m_extraSessions.remove( id );
+    }
+
+    if ( login->session() == m_mainSession ) {
+      m_mainSession = 0;
+    }
+
+    KIMAP::LogoutJob *logout = new KIMAP::LogoutJob( login->session() );
+    QObject::connect( logout, SIGNAL( result( KJob* ) ),
+                      login->session(), SLOT( deleteLater() ) );
+    logout->start();
+
   }
 }
 
 void ImapAccount::onCapabilitiesTestDone( KJob *job )
 {
+  KIMAP::CapabilitiesJob *capJob = qobject_cast<KIMAP::CapabilitiesJob*>( job );
+
   if ( job->error() ) {
-    emit error( CapabilitiesTestError,
+    emit error( capJob->session(),
+                CapabilitiesTestError,
                 i18n( "Could not test the capabilities supported by the IMAP server %1.", m_server ) );
     disconnect();
     return;
   }
 
-  KIMAP::CapabilitiesJob *capJob = qobject_cast<KIMAP::CapabilitiesJob*>( job );
   m_capabilities = capJob->capabilities();
   QStringList expected;
   expected << "IMAP4REV1";
 
-  if ( !m_capabilities.contains( "X-GM-EXT-1" ) ) {
+  // Both GMail and GMX servers seem to be lying about their capabilities
+  // They don't report UIDPLUS correctly so don't check for it explicitly
+  // if it's one of those servers.
+  if ( !m_capabilities.contains( "X-GM-EXT-1" )
+    && !capJob->session()->serverGreeting().contains( "GMX" ) ) {
     expected << "UIDPLUS";
   }
 
@@ -263,21 +319,53 @@ void ImapAccount::onCapabilitiesTestDone( KJob *job )
   }
 
   if ( !missing.isEmpty() ) {
-    emit error( IncompatibleServerError,
+    emit error( capJob->session(),
+                IncompatibleServerError,
                 i18n( "Cannot use the IMAP server %1, some mandatory capabilities are missing: %2. "
                       "Please ask your sysadmin to upgrade the server.", m_server, missing.join( ", " ) ) );
     disconnect();
     return;
   }
 
-  emit success();
+  // If the extension is supported, grab the namespaces from the server
+  if ( m_capabilities.contains( "NAMESPACE" ) ) {
+    KIMAP::NamespaceJob *nsJob = new KIMAP::NamespaceJob( m_mainSession );
+    QObject::connect( nsJob, SIGNAL( result( KJob* ) ), SLOT( onNamespacesTestDone( KJob* ) ) );
+    nsJob->start();
+    return;
+  } else {
+    emit success( capJob->session() );
+  }
+}
+
+void ImapAccount::onNamespacesTestDone( KJob *job )
+{
+  KIMAP::NamespaceJob *nsJob = qobject_cast<KIMAP::NamespaceJob*>( job );
+
+  if ( nsJob->containsEmptyNamespace() ) {
+    // When we got the empty namespace here, we assume that the other
+    // ones can be freely ignored and that the server will give us all
+    // the mailboxes if we list from the empty namespace itself...
+
+    m_namespaces.clear();
+
+  } else {
+    // ... otherwise we assume that we have to list explicitly each
+    // namespace
+
+    m_namespaces = nsJob->personalNamespaces()
+                 + nsJob->userNamespaces()
+                 + nsJob->sharedNamespaces();
+  }
+
+  emit success( nsJob->session() );
 }
 
 /******************* Private ***********************************************/
 
 bool ImapAccount::connect( const QString &password )
 {
-  if ( m_session==0 ) {
+  if ( m_mainSession==0 ) {
     doConnect( password );
     return true;
   } else {
@@ -285,17 +373,20 @@ bool ImapAccount::connect( const QString &password )
   }
 }
 
-bool ImapAccount::disconnect()
+void ImapAccount::disconnect()
 {
-  if ( m_session==0 ) {
-    return false;
-  } else {
-    KIMAP::LogoutJob *logout = new KIMAP::LogoutJob( m_session );
-    QObject::connect( logout, SIGNAL( result( KJob* ) ), m_session, SLOT( deleteLater() ) );
+  QList<KIMAP::Session*> sessions;
+  sessions << m_mainSession;
+  sessions+= m_extraSessions.values();
+
+  foreach ( KIMAP::Session *session, sessions ) {
+    KIMAP::LogoutJob *logout = new KIMAP::LogoutJob( session );
+    QObject::connect( logout, SIGNAL( result( KJob* ) ), session, SLOT( deleteLater() ) );
     logout->start();
-    m_session = 0;
-    return true;
   }
+
+  m_mainSession = 0;
+  m_extraSessions.clear();
 }
 
 #include "imapaccount.moc"
